import numpy as np
import pandas as pd
import pytest

from vivarium.framework.results.manager import ResultsManager
from vivarium.framework.results.stratification import Stratification

from .mocks import (
    CATEGORIES,
    NAME,
    SOURCES,
    STUDENT_HOUSES,
    STUDENT_TABLE,
    sorting_hat_bad_mapping,
    sorting_hat_serial,
    sorting_hat_vector,
)


#########
# Tests #
#########
@pytest.mark.parametrize(
    "name, sources, categories, mapper, is_vectorized, expected_output",
    [
        (  # expected output for vectorized
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_vector,
            True,
            STUDENT_HOUSES,
        ),
        (  # expected output for non-vectorized
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_serial,
            False,
            STUDENT_HOUSES,
        ),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_stratification(name, sources, categories, mapper, is_vectorized, expected_output):
    my_stratification = Stratification(name, sources, categories, mapper, is_vectorized)
    output = my_stratification(STUDENT_TABLE)[name]
    assert output.eq(expected_output).all()


@pytest.mark.parametrize(
    "name, sources, categories, mapper, is_vectorized, expected_exception",
    [
        (  # empty sources list with no defined mapper (default mapper)
            NAME,
            [],
            CATEGORIES,
            None,
            True,
            ValueError,
        ),
        (  # sources list with more than one column with no defined mapper (default mapper)
            NAME,
            SOURCES,
            CATEGORIES,
            None,
            True,
            ValueError,
        ),
        (  # empty sources list with no defined mapper (default mapper)
            NAME,
            [],
            CATEGORIES,
            None,
            True,
            ValueError,
        ),
        (  # empty categories list
            NAME,
            SOURCES,
            [],
            None,
            True,
            ValueError,
        ),
    ],
)
def test_stratification_init_raises(
    name, sources, categories, mapper, is_vectorized, expected_exception
):
    with pytest.raises(expected_exception):
        assert Stratification(name, sources, categories, mapper, is_vectorized)


@pytest.mark.parametrize(
    "name, sources, categories, mapper, is_vectorized, expected_exception",
    [
        (
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_bad_mapping,
            False,
            ValueError,
        ),
        (
            NAME,
            ["middle_initial"],
            CATEGORIES,
            sorting_hat_vector,
            True,
            KeyError,
        ),
        (
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_serial,
            True,
            Exception,
        ),
        (
            NAME,
            SOURCES,
            CATEGORIES,
            sorting_hat_vector,
            False,
            Exception,
        ),
    ],
    ids=[
        "category_not_in_categories",
        "source_not_in_population_columns",
        "vectorized_with_serial_mapper",
        "not_vectorized_with_serial_mapper",
    ],
)
def test_stratification_call_raises(
    name, sources, categories, mapper, is_vectorized, expected_exception
):
    my_stratification = Stratification(name, sources, categories, mapper, is_vectorized)
    with pytest.raises(expected_exception):
        raise my_stratification(STUDENT_TABLE)


@pytest.mark.parametrize("default_stratifications", [["age", "sex"], ["age"], []])
def test_setting_default_stratifications(default_stratifications, mocker):
    """Test that default stratifications are set as expected."""
    mgr = ResultsManager()
    builder = mocker.Mock()
    builder.configuration.stratification.default = default_stratifications

    mgr.setup(builder)

    assert mgr._results_context.default_stratifications == default_stratifications



from pathlib import Path
from unittest.mock import call

import pytest

from vivarium.framework.artifact.artifact import (
    Artifact,
    ArtifactException,
    _parse_draw_filters,
    _to_tree,
)
from vivarium.framework.artifact.hdf import EntityKey


@pytest.fixture()
def keys_mock():
    keys = [
        "metadata.locations",
        "metadata.keyspace",
        "metadata.versions",
        "cause.all_causes.cause_specific_mortality",
        "cause.all_causes.restrictions",
        "cause.diarrheal_diseases.cause_specific_mortality",
        "cause.diarrheal_diseases.death",
        "cause.diarrheal_diseases.etiologies",
        "cause.diarrheal_diseases.excess_mortality",
        "cause.diarrheal_diseases.incidence",
        "cause.diarrheal_diseases.population_attributable_fraction",
        "cause.diarrheal_diseases.prevalence",
        "cause.diarrheal_diseases.remission",
        "cause.diarrheal_diseases.restrictions",
        "cause.diarrheal_diseases.sequelae",
        "covariate.dtp3_coverage_proportion.estimate",
        "dimensions.full_space",
        "etiology.adenovirus.population_attributable_fraction",
        "etiology.aeromonas.population_attributable_fraction",
        "etiology.amoebiasis.population_attributable_fraction",
        "population.age_bins",
        "population.structure",
        "population.theoretical_minimum_risk_life_expectancy",
        "risk_factor.child_stunting.affected_causes",
        "risk_factor.child_stunting.distribution",
        "risk_factor.child_stunting.exposure",
        "risk_factor.child_stunting.levels",
        "risk_factor.child_stunting.relative_risk",
        "risk_factor.child_stunting.restrictions",
        "sequela.moderate_diarrheal_diseases.disability_weight",
        "sequela.moderate_diarrheal_diseases.healthstate",
        "sequela.moderate_diarrheal_diseases.incidence",
        "sequela.moderate_diarrheal_diseases.prevalence",
        "no_data.key",
    ]
    return keys


@pytest.fixture()
def hdf_mock(mocker, keys_mock):
    mock = mocker.patch("vivarium.framework.artifact.artifact.hdf")

    def mock_load(_, key, __, ___):
        if key in keys_mock:
            if key == "metadata.keyspace":
                return keys_mock
            elif key != "no_data.key":
                return "data"
            else:
                return None

    mock.load.side_effect = mock_load
    mock.get_keys.return_value = keys_mock
    mock.touch.side_effect = lambda _: None
    return mock


@pytest.fixture(scope="function")
def artifact_path(tmp_path):
    ap = tmp_path / "artifact.hdf"
    ap.touch()
    return ap


# keys in test artifact
_KEYS = [
    "population.age_bins",
    "population.structure",
    "population.theoretical_minimum_risk_life_expectancy",
    "cause.all_causes.restrictions",
]


def test_artifact_creation(hdf_mock, keys_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]

    a = Artifact(artifact_path)

    assert a.filter_terms is None
    assert a._cache == {}
    assert a.keys == keys_mock
    hdf_mock.load.called_once_with("metadata.keyspace")

    a = Artifact(artifact_path, filter_terms)

    assert a.path == str(artifact_path)
    assert a.filter_terms == filter_terms
    assert a._cache == {}
    assert a.keys == keys_mock
    hdf_mock.load.called_once_with("metadata.keyspace")


def test_artifact_load_missing_key(hdf_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "not.a_real.key"

    a = Artifact(artifact_path, filter_terms)
    hdf_mock.load.called_once_with("metadata.keyspace")
    hdf_mock.load.reset_mock()
    with pytest.raises(ArtifactException) as err_info:
        a.load(key)

    assert f"{key} should be in {artifact_path}." == str(err_info.value)
    hdf_mock.load.assert_not_called()
    assert a._cache == {}


def test_artifact_load_key_has_no_data(hdf_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "no_data.key"

    a = Artifact(artifact_path, filter_terms)

    with pytest.raises(AssertionError) as err_info:
        a.load(key)

    assert f"Data for {key} is not available. Check your model specification." == str(
        err_info.value
    )
    assert hdf_mock.load.called_once_with(artifact_path, key, filter_terms)
    assert a._cache == {}


def test_artifact_load(hdf_mock, keys_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]

    a = Artifact(artifact_path, filter_terms)
    keys_without_metadata = set(keys_mock) - {
        "metadata.locations",
        "metadata.keyspace",
        "metadata.versions",
    }
    for key in keys_without_metadata:
        if key == "no_data.key":
            continue

        assert key not in a._cache

        result = a.load(key)

        assert hdf_mock.load.called_once_with(artifact_path, key, filter_terms)
        assert key in a._cache
        assert a._cache[key] == "data"
        assert result == "data"

        hdf_mock.load.reset_mock()


def test_artifact_contains(hdf_mock, keys_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    not_a_key = "not.a.key"

    art = Artifact(artifact_path, filter_terms)

    for key in art.keys:
        assert key in art

    for art_key in art:
        assert art_key in art.keys

    assert not_a_key not in art
    assert not_a_key not in art.keys


def test_artifact_write_duplicate_key(hdf_mock, keys_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "population.structure"

    art = Artifact(artifact_path, filter_terms)
    initial_keys = art.keys
    assert initial_keys == keys_mock

    with pytest.raises(ArtifactException) as err_info:
        art.write(key, "data")

    assert f"{key} already in artifact." == str(err_info.value)
    assert key in art
    assert key not in art._cache
    hdf_mock.write.called_once_with(artifact_path, "metadata.keyspace", ["metadata.keyspace"])
    hdf_mock.remove.assert_not_called()
    assert art.keys == initial_keys


def test_artifact_write_no_data(hdf_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "new.key"

    a = Artifact(artifact_path, filter_terms)
    initial_keys = a.keys

    assert key not in a

    with pytest.raises(ArtifactException):
        a.write(key, None)

    assert key not in a
    assert key not in a._cache
    hdf_mock.write.called_once_with(artifact_path, "metadata.keyspace", ["metadata.keyspace"])
    hdf_mock.remove.assert_not_called()
    assert a.keys == initial_keys


def test_artifact_write(hdf_mock, keys_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "new.key"

    a = Artifact(artifact_path, filter_terms)
    initial_keys = a.keys

    assert key not in a

    a.write(key, "data")

    assert key in a
    assert key not in a._cache
    expected_call = [
        call(artifact_path, key, "data"),
        call(artifact_path, "metadata.keyspace", keys_mock + [key]),
    ]
    assert hdf_mock.write.call_args_list == expected_call
    assert set(a.keys) == set(initial_keys + [key])


def test_artifact_write_and_load_with_different_key_types(hdf_mock, keys_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    string_key = "new.key"
    entity_key = EntityKey(string_key)

    key_pairs = {string_key: entity_key, entity_key: string_key}

    for write_key, load_key in key_pairs.items():
        a = Artifact(artifact_path, filter_terms)

        assert write_key not in a
        a.write(write_key, "data")

        keys_mock.append(write_key)
        assert load_key not in a._cache

        a.load(load_key)

        assert hdf_mock.load.called_once_with(artifact_path, load_key, filter_terms)
        assert load_key in a._cache

        hdf_mock.load.reset_mock()
        keys_mock.remove(write_key)


def test_artifact_write_and_reopen_then_load_with_entity_key(
    hdf_mock, keys_mock, artifact_path
):
    filter_terms = ["location == Global", "draw == 10"]
    key = EntityKey("new.key")

    a = Artifact(artifact_path, filter_terms)

    assert key not in a
    a.write(key, "data")

    keys_mock.append(key)

    a_again = Artifact(artifact_path, filter_terms)

    assert key not in a_again._cache

    a_again.load(key)

    assert hdf_mock.load.called_once_with(artifact_path, key, filter_terms)
    assert key in a_again._cache

    hdf_mock.load.reset_mock()
    keys_mock.remove(key)


def test_remove_bad_key(hdf_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "non_existent.key"
    a = Artifact(artifact_path, filter_terms)
    initial_keys = a.keys

    assert key not in a
    assert key not in a._cache

    with pytest.raises(ArtifactException) as err_info:
        a.remove(key)

    assert f"Trying to remove non-existent key {key} from artifact." == str(err_info.value)
    assert key not in a
    assert key not in a._cache
    hdf_mock.remove.assert_not_called()
    hdf_mock.write.called_once_with(artifact_path, "metadata.keyspace", ["metadata.keyspace"])
    assert a.keys == initial_keys


def test_remove_no_cache(hdf_mock, keys_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "population.structure"

    a = Artifact(artifact_path, filter_terms)

    initial_keys = a.keys[:]

    assert key in initial_keys
    assert key not in a._cache

    a.remove(key)

    assert key not in a
    assert key not in a._cache
    assert set(initial_keys).difference(a.keys) == {key}
    expected_calls_remove = [
        call(artifact_path, "metadata.keyspace"),
        call(artifact_path, key),
    ]
    assert hdf_mock.remove.call_args_list == expected_calls_remove
    expected_calls_write = [
        call(artifact_path, "metadata.keyspace", [k for k in keys_mock if k != key]),
    ]
    assert hdf_mock.write.call_args_list == expected_calls_write


def test_remove(hdf_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "population.structure"

    a = Artifact(artifact_path, filter_terms)
    a._cache[key] = "data"

    assert key in a
    assert key in a._cache

    a.remove(key)

    assert key not in a
    assert key not in a._cache

    expected_calls = [call(artifact_path, "metadata.keyspace"), call(artifact_path, key)]
    assert hdf_mock.remove.call_args_list == expected_calls


def test_clear_cache(hdf_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "population.structure"

    a = Artifact(artifact_path, filter_terms)
    a.clear_cache()

    assert a._cache == {}

    a._cache[key] = "data"
    a.clear_cache()

    assert a._cache == {}


def test_loading_key_leaves_filters_unchanged(hdf_mock, artifact_path):
    # loading each key will drop the fake_filter from filter_terms for that key
    # make sure that artifact's filter terms stay the same though
    filter_terms = ["location == Global", "draw == 10", "fake_filter"]

    a = Artifact(artifact_path, filter_terms=filter_terms)

    for key in _KEYS:
        a.load(key)
        assert a.filter_terms == filter_terms


def test_replace(hdf_mock, keys_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "new.key"

    a = Artifact(artifact_path, filter_terms=filter_terms)

    assert key not in a

    a.write(key, "data")
    keyspace_key = "metadata.keyspace"
    new_keyspace = [k for k in keys_mock + [key]]

    assert hdf_mock.write.call_args_list == [
        call(artifact_path, key, "data"),
        call(artifact_path, keyspace_key, new_keyspace),
    ]

    hdf_mock.reset_mock()

    a.replace(key, "new_data")

    # keyspace will be remove first in self.remove from a.replace then self.write from a.replace
    expected_calls_remove = [
        call(artifact_path, keyspace_key),
        call(artifact_path, key),
        call(artifact_path, keyspace_key),
    ]
    assert hdf_mock.remove.call_args_list == expected_calls_remove

    expected_calls_write = [
        call(artifact_path, keyspace_key, new_keyspace),
        call(artifact_path, key, "new_data"),
        call(artifact_path, keyspace_key, new_keyspace),
    ]
    assert hdf_mock.write.call_args_list == expected_calls_write
    assert key in a


def test_replace_nonexistent_key(hdf_mock, artifact_path):
    filter_terms = ["location == Global", "draw == 10"]
    key = "new.key"

    a = Artifact(artifact_path, filter_terms=filter_terms)
    hdf_mock.called_once_with(key)
    assert key not in a

    hdf_mock.reset_mock()
    with pytest.raises(ArtifactException):
        a.replace(key, "new_data")

    hdf_mock.write.assert_not_called()
    hdf_mock.remove.assert_not_called()


def test_to_tree():
    keys = [
        "population.structure",
        "population.age_groups",
        "cause.diarrheal_diseases.incidence",
        "cause.diarrheal_diseases.prevalence",
    ]

    key_tree = {
        "population": {
            "structure": [],
            "age_groups": [],
        },
        "cause": {"diarrheal_diseases": ["incidence", "prevalence"]},
    }

    assert _to_tree(keys) == key_tree


def test_create_hdf(tmpdir):
    path = Path(tmpdir) / "test.hdf"
    assert not path.is_file()

    with pytest.warns(UserWarning, match="No artifact found"):
        test_artifact = Artifact(path)
    assert path.is_file()
    assert "metadata.keyspace" in test_artifact

    test_artifact.write("new.key", "data")
    assert "new.key" in test_artifact

    #  check whether the existing path was NOT wiped out
    new_artifact = Artifact(test_artifact.path)
    assert new_artifact.path == test_artifact.path
    assert "new.key" in new_artifact


def test_keys_initialization(tmpdir):
    path = Path(tmpdir) / "test.hdf"

    with pytest.warns(UserWarning, match="No artifact found"):
        test_artifact = Artifact(path)
    test_key = test_artifact._keys

    assert test_artifact._path == test_key._path
    assert test_key._keys == ["metadata.keyspace"]

    test_artifact.write("new.keys", "data")
    assert test_key._keys == ["metadata.keyspace", "new.keys"]
    assert test_key.to_list() == test_artifact.keys


def test_keys_append(tmpdir):
    path = Path(tmpdir) / "test.hdf"
    with pytest.warns(UserWarning, match="No artifact found"):
        test_artifact = Artifact(path)
    test_keys = test_artifact._keys

    test_artifact.write("test.keys", "data")
    assert "test.keys" in test_artifact
    assert "test.keys" in test_keys
    assert (
        test_keys._keys
        == ["metadata.keyspace", "test.keys"]
        == [str(k) for k in test_artifact.keys]
    )


def test_keys_remove(tmpdir):
    path = Path(tmpdir) / "test.hdf"
    with pytest.warns(UserWarning, match="No artifact found"):
        test_artifact = Artifact(path)
    test_keys = test_artifact._keys

    test_artifact.write("test.keys1", "data")
    test_artifact.write("test.keys2", "data")
    assert "test.keys1" in test_artifact and "test.keys2" in test_artifact
    assert "test.keys1" in test_keys and "test.keys2" in test_keys

    test_artifact.remove("test.keys2")
    assert "test.keys1" in test_artifact and "test.keys2" not in test_artifact
    assert "test.keys1" in test_keys and "test.keys2" not in test_keys


@pytest.mark.parametrize(
    "filters, error, match",
    [
        (
            ["draw == 0 & year > 2011", "age < 2015 | draw in [1, 2, 3]"],
            ValueError,
            "only supply one",
        ),
        (
            ['year > 2010 & age < 5 & parameter == "cat1" & ' "draw==0 | draw==100"],
            ValueError,
            "only supply one",
        ),
        (["draw >= 10"], NotImplementedError, "only supported"),
        (["draw<5"], NotImplementedError, "only supported"),
    ],
)
def test__parse_draw_filters_fail(filters, error, match):
    with pytest.raises(error, match=match):
        _parse_draw_filters(filters)


@pytest.mark.parametrize(
    "draw_operator, draw_values",
    [
        ("=", [5]),
        ("==", [10]),
        (" = ", [100]),
        (" == ", [12]),
        (" in ", [1, 7, 160]),
        ("    in            ", [140, 2, 14]),
    ],
)
def test__parse_draw_filters_pass(draw_operator, draw_values):
    draw_filter = (
        f'draw{draw_operator}{draw_values if "in" in draw_operator else draw_values[0]}'
    )
    expected_cols = [f"draw_{d}" for d in draw_values] + ["value"]

    assert _parse_draw_filters([draw_filter]) == expected_cols

    complicated_filter = [
        f"year > 2010 & age < 12",
        "other_col in [1, 2, 40]",
        "one_more > 10",
    ]
    for i in range(len(complicated_filter)):
        filters = complicated_filter.copy()
        filters[i] += " | " + draw_filter
        assert _parse_draw_filters(filters) == expected_cols

    assert _parse_draw_filters(complicated_filter + [draw_filter]) == expected_cols

    assert _parse_draw_filters([]) is None



"""
==========================
Vivarium Testing Utilities
==========================

Utility functions and classes to make testing ``vivarium`` components easier.

"""
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd

from vivarium import Component
from vivarium.framework import randomness
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.population import SimulantData
from vivarium.framework.randomness.index_map import IndexMap


class NonCRNTestPopulation(Component):
    CONFIGURATION_DEFAULTS = {
        "population": {
            "age_start": 0,
            "age_end": 100,
            "exit_age": None,
        },
    }

    @property
    def columns_created(self) -> List[str]:
        return ["age", "sex", "location", "alive", "entrance_time", "exit_time"]

    def setup(self, builder: Builder) -> None:
        self.config = builder.configuration
        self.randomness = builder.randomness.get_stream(
            "population_age_fuzz", initializes_crn_attributes=True
        )

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        age_start = pop_data.user_data.get("age_start", self.config.population.age_start)
        age_end = pop_data.user_data.get("age_end", self.config.population.age_end)
        location = self.config.input_data.location

        population = _non_crn_build_population(
            pop_data.index,
            age_start,
            age_end,
            location,
            pop_data.creation_time,
            pop_data.creation_window,
            self.randomness,
        )
        self.population_view.update(population)

    def on_time_step(self, event: Event) -> None:
        population = self.population_view.get(event.index, query="alive == 'alive'")
        population["age"] += event.step_size / pd.Timedelta(days=365)
        self.population_view.update(population)


class TestPopulation(NonCRNTestPopulation):
    def setup(self, builder: Builder) -> None:
        super().setup(builder)
        self.age_randomness = builder.randomness.get_stream(
            "age_initialization", initializes_crn_attributes=True
        )
        self.register = builder.randomness.register_simulants

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        age_start = pop_data.user_data.get("age_start", self.config.population.age_start)
        age_end = pop_data.user_data.get("age_end", self.config.population.age_end)
        age_draw = self.age_randomness.get_draw(pop_data.index)
        if age_start == age_end:
            age = age_draw * (pop_data.creation_window / pd.Timedelta(days=365)) + age_start
        else:
            age = age_draw * (age_end - age_start) + age_start

        core_population = pd.DataFrame(
            {"entrance_time": pop_data.creation_time, "age": age.values}, index=pop_data.index
        )
        self.register(core_population)

        location = (
            self.config.input_data.location
            if "location" in self.config.input_data.keys()
            else None
        )
        population = _build_population(core_population, location, self.randomness)
        self.population_view.update(population)


def _build_population(core_population, location, randomness_stream):
    index = core_population.index

    population = pd.DataFrame(
        {
            "age": core_population["age"],
            "entrance_time": core_population["entrance_time"],
            "sex": randomness_stream.choice(
                index, ["Male", "Female"], additional_key="sex_choice"
            ),
            "alive": pd.Series("alive", index=index),
            "location": location,
            "exit_time": pd.NaT,
        },
        index=index,
    )
    return population


def _non_crn_build_population(
    index, age_start, age_end, location, creation_time, creation_window, randomness_stream
):
    if age_start == age_end:
        age = (
            randomness_stream.get_draw(index) * (creation_window / pd.Timedelta(days=365))
            + age_start
        )
    else:
        age = randomness_stream.get_draw(index) * (age_end - age_start) + age_start

    population = pd.DataFrame(
        {
            "age": age,
            "sex": randomness_stream.choice(
                index, ["Male", "Female"], additional_key="sex_choice"
            ),
            "alive": pd.Series("alive", index=index),
            "location": location,
            "entrance_time": creation_time,
            "exit_time": pd.NaT,
        },
        index=index,
    )
    return population


def build_table(value, year_start, year_end, columns=("age", "year", "sex", "value")):
    value_columns = columns[3:]
    if not isinstance(value, list):
        value = [value] * len(value_columns)

    if len(value) != len(value_columns):
        raise ValueError("Number of values must match number of value columns")

    rows = []
    for age in range(0, 140):
        for year in range(year_start, year_end + 1):
            for sex in ["Male", "Female"]:
                r_values = []
                for v in value:
                    if v is None:
                        r_values.append(np.random.random())
                    elif callable(v):
                        r_values.append(v(age, sex, year))
                    else:
                        r_values.append(v)
                rows.append([age, age + 1, year, year + 1, sex] + r_values)
    return pd.DataFrame(
        rows,
        columns=["age_start", "age_end", "year_start", "year_end", "sex"]
        + list(value_columns),
    )


def make_dummy_column(name, initial_value):
    class DummyColumnMaker:
        @property
        def name(self):
            return "dummy_column_maker"

        def setup(self, builder):
            self.population_view = builder.population.get_view([name])
            builder.population.initializes_simulants(self.make_column, creates_columns=[name])

        def make_column(self, pop_data):
            self.population_view.update(
                pd.Series(initial_value, index=pop_data.index, name=name)
            )

        def __repr__(self):
            return f"dummy_column(name={name}, initial_value={initial_value})"

    return DummyColumnMaker()


def get_randomness(
    key="test",
    clock=lambda: pd.Timestamp(1990, 7, 2),
    seed=12345,
    initializes_crn_attributes=False,
):
    return randomness.RandomnessStream(
        key,
        clock,
        seed=seed,
        index_map=IndexMap(),
        initializes_crn_attributes=initializes_crn_attributes,
    )


def reset_mocks(mocks):
    for mock in mocks:
        mock.reset_mock()


def metadata(file_path, layer="override"):
    return {"layer": layer, "source": str(Path(file_path).resolve())}



from pathlib import Path

import pytest
import tables
from _pytest.logging import LogCaptureFixture
from loguru import logger

from vivarium.framework.configuration import (
    build_model_specification,
    build_simulation_configuration,
)
from vivarium.testing_utilities import metadata


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture
def caplog(caplog: LogCaptureFixture):
    handler_id = logger.add(caplog.handler, format="{message}")
    yield caplog
    logger.remove(handler_id)


@pytest.fixture
def base_config():
    config = build_simulation_configuration()
    config.update(
        {
            "time": {
                "start": {
                    "year": 1990,
                },
                "end": {"year": 2010},
                "step_size": 30.5,
            },
            "randomness": {"key_columns": ["entrance_time", "age"]},
        },
        **metadata(__file__, layer="model_override"),
    )
    return config


@pytest.fixture
def test_data_dir():
    data_dir = Path(__file__).resolve().parent / "test_data"
    assert data_dir.exists(), "Test directory structure is broken"
    return data_dir


@pytest.fixture(params=[".yaml", ".yml"])
def test_spec(request, test_data_dir):
    return test_data_dir / f"mock_model_specification{request.param}"


@pytest.fixture(params=[".yaml", ".yml"])
def test_user_config(request, test_data_dir):
    return test_data_dir / f"mock_user_config{request.param}"


@pytest.fixture
def model_specification(mocker, test_spec, test_user_config):
    expand_user_mock = mocker.patch("vivarium.framework.configuration.Path.expanduser")
    expand_user_mock.return_value = test_user_config
    return build_model_specification(test_spec)


@pytest.fixture
def hdf_file_path(tmpdir, test_data_dir):
    """This file contains the following:
    Object Tree:
        / (RootGroup) ''
        /cause (Group) ''
        /population (Group) ''
        /population/age_bins (Group) ''
        /population/age_bins/table (Table(23,), shuffle, zlib(9)) ''
        /population/structure (Group) ''
        /population/structure/table (Table(1863,), shuffle, zlib(9)) ''
        /population/theoretical_minimum_risk_life_expectancy (Group) ''
        /population/theoretical_minimum_risk_life_expectancy/table (Table(10502,), shuffle, zlib(9)) ''
        /population/structure/meta (Group) ''
        /population/structure/meta/values_block_1 (Group) ''
        /population/structure/meta/values_block_1/meta (Group) ''
        /population/structure/meta/values_block_1/meta/table (Table(3,), shuffle, zlib(9)) ''
        /cause/all_causes (Group) ''
        /cause/all_causes/restrictions (EArray(166,)) ''
    """
    # Make temporary copy of file for test.
    p = tmpdir.join("artifact.hdf")
    with tables.open_file(str(test_data_dir / "artifact.hdf")) as file:
        file.copy_file(str(p), overwrite=True)
    return p


@pytest.fixture
def hdf_file(hdf_file_path):
    with tables.open_file(str(hdf_file_path)) as file:
        yield file



from typing import TYPE_CHECKING, Callable, List

import pandas as pd

if TYPE_CHECKING:
    # cyclic import
    from vivarium.framework.results.manager import ResultsManager


class ResultsInterface:
    """Builder interface for the results management system.

    The results management system allows users to delegate results production
    to the simulation framework. This process attempts to roughly mimic the
    groupby-apply logic commonly done when manipulating :mod:`pandas`
    DataFrames. The representation of state in the simulation is complex,
    however, as it includes information both in the population state table
    and dynamically generated information available from the
    :class:`value pipelines <vivarium.framework.values.Pipeline>`.
    Additionally, good encapsulation of simulation logic typically has
    results production separated from the modeling code into specialized
    `Observer` components. This often highlights the need for transformations
    of the simulation state into representations that aren't needed for
    modeling, but are required for the stratification of produced results.

    The purpose of this interface is to provide controlled access to a results
    backend by means of the builder object. It exposes methods
    to register stratifications, set default stratifications, and register
    results producers. There is a special case for stratifications generated
    by binning continuous data into categories.

    The expected use pattern would be for a single component to register all
    stratifications required by the model using :func:`register_default_stratifications`,
    :func:`register_stratification`, and :func:`register_binned_stratification`
    as necessary. A “binned stratification” is a stratification special case for
    the very common situation when a single continuous value needs to be binned into
    categorical bins. The `is_vectorized` argument should be True if the mapper
    function expects a DataFrame corresponding to the whole population, and False
    if it expects a row of the DataFrame corresponding to a single simulant.
    """

    def __init__(self, manager: "ResultsManager") -> None:
        self._manager: "ResultsManager" = manager
        self._name = "results_interface"

    @property
    def name(self) -> str:
        """The name of this ResultsInterface."""
        return self._name

    def set_default_stratifications(self, default_stratifications: List[str]):
        self._manager.set_default_stratifications(default_stratifications)

    # TODO: It is not reflected in the sample code here, but the “when” parameter should be added
    #  to the stratification registration calls, probably as a List. Consider this after observer implementation
    def register_stratification(
        self,
        name: str,
        categories: List[str],
        mapper: Callable = None,
        is_vectorized: bool = False,
        requires_columns: List[str] = (),
        requires_values: List[str] = (),
    ) -> None:
        """Register quantities to observe.

        Parameters
        ----------
        name
            Name of the of the column created by the stratification.
        categories
            List of string values that the mapper is allowed to output.
        mapper
            A callable that emits values in `categories` given inputs from columns
            and values in the `requires_columns` and `requires_values`, respectively.
        is_vectorized
            `True` if the mapper function expects a `DataFrame`, and `False` if it
            expects a row of the `DataFrame` and should be used by calling :func:`df.apply`.
        requires_columns
            A list of the state table columns that already need to be present
            and populated in the state table before the pipeline modifier
            is called.
        requires_values
            A list of the value pipelines that need to be properly sourced
            before the pipeline modifier is called.

        Returns
        ------
        None
        """
        self._manager.register_stratification(
            name,
            categories,
            mapper,
            is_vectorized,
            requires_columns,
            requires_values,
        )

    def register_binned_stratification(
        self,
        target: str,
        binned_column: str,
        bins: List = (),
        labels: List[str] = (),
        target_type: str = "column",
        **cut_kwargs,
    ) -> None:
        """Register a continuous `target` quantity to observe into bins in a `binned_column`.

        Parameters
        ----------
        target
            String name of the state table column or value pipeline used to stratify.
        binned_column
            String name of the column for the binned quantities.
        bins
            List of scalars defining the bin edges, passed to :meth: pandas.cut. Lists
            `bins` and `labels` must be of equal length.
        labels
            List of string labels for bins. Lists `bins` and `labels` must be of equal length.
        target_type
            "column" or "value"
        **cut_kwargs
            Keyword arguments for :meth: pandas.cut.

        Returns
        ------
        None
        """
        self._manager.register_binned_stratification(
            target, target_type, binned_column, bins, labels, **cut_kwargs
        )

    def register_observation(
        self,
        name: str,
        pop_filter: str = "tracked==True",
        aggregator_sources: List[str] = None,
        aggregator: Callable[[pd.DataFrame], float] = len,
        requires_columns: List[str] = (),
        requires_values: List[str] = (),
        additional_stratifications: List[str] = (),
        excluded_stratifications: List[str] = (),
        when: str = "collect_metrics",
    ) -> None:
        """Provide the results system all the information it needs to perform the observation.

        Parameters
        ----------
        name
            String name for the observation.
        pop_filter
            A Pandas query filter string to filter the population down to the simulants who should
            be considered for the observation.
        aggregator_sources
            A list of population view columns to be used in the aggregator.
        aggregator
            A function that computes the quantity for the observation.
        requires_columns
            A list of the state table columns that are required by either the pop_filter or the aggregator.
        requires_values
            A list of the value pipelines that are required by either the pop_filter or the aggregator.
        additional_stratifications
            A list of additional :class:`stratification <vivarium.framework.results.stratification.Stratification>`
            names by which to stratify.
        excluded_stratifications
            A list of default :class:`stratification <vivarium.framework.results.stratification.Stratification>`
            names to remove from the observation.
        when
            String name of the phase of a time-step the observation should happen. Valid values are:
            `"time_step__prepare"`, `"time_step"`, `"time_step__cleanup"`, `"collect_metrics"`.

        Returns
        ------
        None
        """
        self._manager.register_observation(
            name,
            pop_filter,
            aggregator_sources,
            aggregator,
            requires_columns,
            requires_values,
            additional_stratifications,
            excluded_stratifications,
            when,
        )



import numpy as np
import pandas as pd
import pytest

from vivarium.framework.values import ValuesManager, list_combiner, union_post_processor


@pytest.fixture
def manager(mocker):
    manager = ValuesManager()
    builder = mocker.MagicMock()
    builder.step_size = lambda: lambda: pd.Timedelta(days=1)
    manager.setup(builder)
    return manager


def test_replace_combiner(manager):
    value = manager.register_value_producer("test", source=lambda: 1)

    assert value() == 1

    manager.register_value_modifier("test", modifier=lambda v: 42)
    assert value() == 42

    manager.register_value_modifier("test", lambda v: 84)
    assert value() == 84


def test_joint_value(manager):
    # This is the normal configuration for PAF and disability weight type values
    index = pd.Index(range(10))

    value = manager.register_value_producer(
        "test",
        source=lambda idx: [pd.Series(0.0, index=idx)],
        preferred_combiner=list_combiner,
        preferred_post_processor=union_post_processor,
    )
    assert np.all(value(index) == 0)

    manager.register_value_modifier("test", modifier=lambda idx: pd.Series(0.5, index=idx))
    assert np.all(value(index) == 0.5)

    manager.register_value_modifier("test", modifier=lambda idx: pd.Series(0.5, index=idx))
    assert np.all(value(index) == 0.75)


def test_contains(manager):
    value = "test_value"
    rate = "test_rate"

    assert value not in manager
    assert rate not in manager

    manager.register_value_producer("test_value", source=lambda: 1)
    assert value in manager
    assert rate not in manager



from typing import Any, Dict

import pytest

from vivarium.framework.components.manager import (
    ComponentConfigError,
    ComponentManager,
    OrderedComponentSet,
)
from vivarium.framework.configuration import build_simulation_configuration

from .mocks import MockComponentA, MockComponentB, MockGenericComponent


def test_component_set_add():
    component_list = OrderedComponentSet()

    component_0 = MockComponentA(name="component_0")
    component_list.add(component_0)

    component_1 = MockComponentA(name="component_1")
    component_list.add(component_1)

    # duplicates by name
    with pytest.raises(ComponentConfigError, match="duplicate name"):
        component_list.add(component_0)


def test_component_set_update():
    component_list = OrderedComponentSet()

    components = [MockComponentA(name="component_0"), MockComponentA("component_1")]

    component_list.update(components)

    with pytest.raises(ComponentConfigError, match="duplicate name"):
        component_list.update(components)


def test_component_set_initialization():
    component_1 = MockComponentA()
    component_2 = MockComponentB()

    component_list = OrderedComponentSet(component_1, component_2)
    assert component_list.components == [component_1, component_2]


def test_component_set_pop():
    component = MockComponentA()
    component_list = OrderedComponentSet(component)

    c = component_list.pop()
    assert c == component

    with pytest.raises(IndexError):
        component_list.pop()


def test_component_set_contains():
    component_list = OrderedComponentSet()

    assert not bool(component_list)
    assert len(component_list) == 0

    component_1 = MockComponentA()
    component_2 = MockComponentB()
    component_3 = MockComponentA(name="absent")
    component_list = OrderedComponentSet(component_1, component_2)

    assert component_1 in component_list
    assert component_3 not in component_list

    with pytest.raises(ComponentConfigError, match="no name"):
        _ = 10 in component_list


def test_component_set_eq():
    component_1 = MockComponentA()
    component_2 = MockComponentB()
    component_list = OrderedComponentSet(component_1, component_2)

    assert component_list == component_list
    assert component_list != 10

    second_list = OrderedComponentSet(component_1)
    assert component_list != second_list


def test_component_set_bool_len():
    component_list = OrderedComponentSet()

    assert not bool(component_list)
    assert len(component_list) == 0

    component_1 = MockComponentA()
    component_2 = MockComponentB()
    component_list = OrderedComponentSet(component_1, component_2)

    assert bool(component_list)
    assert len(component_list) == 2


def test_component_set_dunder_add():
    l1 = OrderedComponentSet(*[MockComponentA(name=str(i)) for i in range(5)])
    l2 = OrderedComponentSet(*[MockComponentA(name=str(i)) for i in range(5, 10)])
    combined = OrderedComponentSet(*[MockComponentA(name=str(i)) for i in range(10)])

    assert l1 + l2 == combined


def test_manager_init():
    m = ComponentManager()
    assert not m._managers
    assert not m._components
    assert not m.configuration
    assert m.name == "component_manager"
    assert repr(m) == str(m) == "ComponentManager()"


def test_manager_get_file():
    class Test:
        pass

    t = Test()
    assert ComponentManager._get_file(t) == __file__
    t.__module__ = "__main__"
    assert ComponentManager._get_file(t) == "__main__"


def test_flatten_simple():
    components = [MockComponentA(name=str(i)) for i in range(10)]
    assert ComponentManager._flatten(components) == components


def test_flatten_with_lists():
    components = []
    for i in range(5):
        for j in range(5):
            components.append(MockComponentA(name=str(5 * i + j)))
    out = ComponentManager._flatten(components)
    expected = [MockComponentA(name=str(i)) for i in range(25)]
    assert out == expected


def test_flatten_with_sub_components():
    components = []
    for i in range(5):
        name, *args = [str(5 * i + j) for j in range(5)]
        components.append(MockComponentB(*args, name=name))
    out = ComponentManager._flatten(components)
    expected = [MockComponentB(name=str(i)) for i in range(25)]
    assert out == expected


def test_flatten_with_nested_sub_components():
    def nest(start, depth):
        if depth == 1:
            return MockComponentA(name=str(start))
        c = MockComponentA(name=str(start))
        c._sub_components = [nest(start + 1, depth - 1)]
        return c

    components = []
    for i in range(5):
        components.append(nest(5 * i, 5))
    out = ComponentManager._flatten(components)
    expected = [MockComponentA(name=str(i)) for i in range(25)]
    assert out == expected

    # Lists with nested subcomponents
    out = ComponentManager._flatten([components, components])
    assert out == 2 * expected


def test_setup_components(mocker):
    builder = mocker.Mock()
    mock_a = MockComponentA("test_a")
    mock_b = MockComponentB("test_b")
    components = OrderedComponentSet(mock_a, mock_b)
    ComponentManager._setup_components(builder, components)

    assert mock_a.builder_used_for_setup is None  # class has no setup method
    assert mock_b.builder_used_for_setup is builder

    builder.value.register_value_modifier.assert_called_once_with("metrics", mock_b.metrics)


def test_apply_configuration_defaults():
    config = build_simulation_configuration()
    c = config.to_dict()
    cm = ComponentManager()
    cm.configuration = config
    component = MockGenericComponent("test_component")

    cm.apply_configuration_defaults(component)
    c.update(component.configuration_defaults)
    assert config.to_dict() == c


def test_apply_configuration_defaults_no_op():
    config = build_simulation_configuration()
    c = config.to_dict()
    cm = ComponentManager()
    cm.configuration = config
    component = MockComponentA()

    cm.apply_configuration_defaults(component)
    assert config.to_dict() == c


def test_apply_configuration_defaults_duplicate():
    config = build_simulation_configuration()

    cm = ComponentManager()
    cm.configuration = config
    component = MockGenericComponent("test_component")

    cm.apply_configuration_defaults(component)
    cm._components.add(component)
    with pytest.raises(ComponentConfigError, match="but it has already been set"):
        cm.apply_configuration_defaults(component)


def test_apply_configuration_defaults_bad_structure():
    class BadConfigComponent(MockComponentA):
        @property
        def configuration_defaults(self) -> Dict[str, Any]:
            return {"test_component": "val"}

    config = build_simulation_configuration()

    cm = ComponentManager()
    cm.configuration = config
    component1 = MockGenericComponent("test_component")
    component2 = BadConfigComponent(name="test_component2")

    cm.apply_configuration_defaults(component1)
    cm._components.add(component1)
    with pytest.raises(ComponentConfigError, match="attempting to alter the structure"):
        cm.apply_configuration_defaults(component2)


def test_add_components():
    config = build_simulation_configuration()
    cm = ComponentManager()
    cm.configuration = config

    assert not cm._managers
    managers = [MockGenericComponent(f"manager_{i}") for i in range(5)]
    components = [MockGenericComponent(f"component_{i}") for i in range(5)]
    cm.add_managers(managers)
    cm.add_components(components)
    assert cm._managers == OrderedComponentSet(*managers)
    assert cm._components == OrderedComponentSet(*components)
    for c in managers + components:
        assert config[c.name].to_dict() == c.configuration_defaults[c.name]

    assert cm.list_components() == {c.name: c for c in components}


@pytest.mark.parametrize(
    "components",
    ([MockComponentA("Eric"), MockComponentB("half", "a", "bee")], [MockComponentA("Eric")]),
)
def test_component_manager_add_components(components):
    config = build_simulation_configuration()
    cm = ComponentManager()
    cm.configuration = config
    cm.add_managers(components)
    assert cm._managers == OrderedComponentSet(*ComponentManager._flatten(components))

    config = build_simulation_configuration()
    cm = ComponentManager()
    cm.configuration = config
    cm.add_components(components)
    assert cm._components == OrderedComponentSet(*ComponentManager._flatten(components))


@pytest.mark.parametrize(
    "components",
    (
        [MockComponentA(), MockComponentA()],
        [MockComponentA(), MockComponentA(), MockComponentB("foo", "bar")],
    ),
)
def test_component_manager_add_components_duplicated(components):
    config = build_simulation_configuration()
    cm = ComponentManager()
    cm.configuration = config
    with pytest.raises(ComponentConfigError, match="duplicate name"):
        cm.add_managers(components)

    config = build_simulation_configuration()
    cm = ComponentManager()
    cm.configuration = config
    with pytest.raises(ComponentConfigError, match="duplicate name"):
        cm.add_components(components)



"""
==============
Output Metrics
==============

Currently, ``vivarium`` uses the :ref:`values pipeline system <values_concept>`
to produce the results, or output metrics, from a simulation.  The metrics
The component here is a normal ``vivarium`` component whose only purpose is
to provide an empty :class:`dict` as the source of the *"Metrics"* pipeline.
It is included by default in all simulations.
"""
from typing import TYPE_CHECKING

from vivarium import Component

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder


class Metrics(Component):
    """This class declares a value pipeline that allows other components to store summary metrics."""

    def setup(self, builder: "Builder") -> None:
        self.metrics = builder.value.register_value_producer(
            "metrics", source=lambda index: {}
        )



from typing import Any, Dict, List, Optional

import pandas as pd

from vivarium import Component
from vivarium.framework.engine import Builder


class Observer(Component):
    ##############
    # Properties #
    ##############

    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        return {
            "mortality": {
                "life_expectancy": 80,
            }
        }

    @property
    def columns_required(self) -> Optional[List[str]]:
        return ["age", "alive"]

    #####################
    # Lifecycle methods #
    #####################

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder) -> None:
        self.life_expectancy = builder.configuration.mortality.life_expectancy
        builder.value.register_value_modifier("metrics", self.metrics)

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def metrics(self, index: pd.Index, metrics: Dict) -> Dict:
        pop = self.population_view.get(index)
        metrics["total_population_alive"] = len(pop[pop.alive == "alive"])
        metrics["total_population_dead"] = len(pop[pop.alive == "dead"])

        metrics["years_of_life_lost"] = (
            self.life_expectancy - pop.age[pop.alive == "dead"]
        ).sum()

        return metrics



from typing import Any, Dict, List, Optional

import pandas as pd
from scipy import spatial

from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.population import SimulantData


class Neighbors(Component):
    ##############
    # Properties #
    ##############
    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        return {"neighbors": {"radius": 10}}

    @property
    def columns_required(self) -> Optional[List[str]]:
        return ["x", "y"]

    #####################
    # Lifecycle methods #
    #####################

    def setup(self, builder: Builder) -> None:
        self.radius = builder.configuration.neighbors.radius

        self.neighbors_calculated = False
        self._neighbors = pd.Series()
        self.neighbors = builder.value.register_value_producer(
            "neighbors", source=self.get_neighbors
        )

    ########################
    # Event-driven methods #
    ########################

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        self._neighbors = pd.Series([[]] * len(pop_data.index), index=pop_data.index)

    def on_time_step(self, event: Event) -> None:
        self.neighbors_calculated = False

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def get_neighbors(self, index: pd.Index) -> pd.Series:
        if not self.neighbors_calculated:
            self._calculate_neighbors()
        return self._neighbors[index]

    ##################
    # Helper methods #
    ##################

    def _calculate_neighbors(self) -> None:
        # Reset our list of neighbors
        pop = self.population_view.get(self._neighbors.index)
        self._neighbors = pd.Series([[]] * len(pop), index=pop.index)

        tree = spatial.KDTree(pop)

        # Iterate over each pair of simulates that are close together.
        for boid_1, boid_2 in tree.query_pairs(self.radius):
            # .iloc is used because query_pairs uses 0,1,... indexing instead of pandas.index
            self._neighbors.iloc[boid_1].append(self._neighbors.index[boid_2])
            self._neighbors.iloc[boid_2].append(self._neighbors.index[boid_1])



"""
=======
Manager
=======

A base Manager class to be used to create manager for use in ``vivarium``
simulations.
"""
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder


class Manager:
    # TODO implement
    def setup(self, builder: "Builder"):
        pass



"""
Integration tests primarily meant to test the CRN guarantees for the Randomness system.

"""
from itertools import cycle
from typing import Iterator, List, Type

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.population import SimulantData
from vivarium.framework.randomness.index_map import IndexMap
from vivarium.framework.randomness.stream import RandomnessStream
from vivarium.interface import InteractiveContext


@pytest.mark.parametrize("initializes_crn_attributes", [True, False])
def test_basic_repeatability(initializes_crn_attributes):
    test_idx = pd.Index(range(100))
    index_map = IndexMap()

    stream_args = {
        "key": "test",
        "clock": lambda: pd.Timestamp("2020-01-01"),
        "seed": "abc",
        "index_map": index_map,
        "initializes_crn_attributes": initializes_crn_attributes,
    }

    stream_base = RandomnessStream(**stream_args)
    draw_base = stream_base.get_draw(test_idx)

    for arg_permutation in [
        {},
        {"key": "test2"},
        {"clock": lambda: pd.Timestamp("2020-01-02")},
        {"seed": "123"},
    ]:
        new_stream_args = {**stream_args, **arg_permutation}
        stream_permutation = RandomnessStream(**new_stream_args)
        draw_1 = stream_permutation.get_draw(test_idx)
        draw_2 = stream_permutation.get_draw(test_idx)

        # Draws with the same RandomnessStream parameterization should be the same
        assert np.allclose(draw_1, draw_2)

        if arg_permutation:
            # Draws from different RandomnessStream parameterizations should be different.
            assert not np.allclose(draw_base, draw_1)


class BasePopulation(Component):
    """
    Population class with parameters to turn on and off CRN and to add simulants on
    time steps.
    """

    @property
    def name(self):
        return "population"

    @property
    def columns_created(self) -> List[str]:
        return ["crn_attr1", "crn_attr2", "other_attr1"]

    def __init__(self, with_crn: bool, sims_to_add: Iterator = cycle([0])):
        """
        Parameters
        ----------
        with_crn
            Boolean switch to turn on CRN implementation
        sims_to_add
            An iterator that yields a number of simulants to add on the
            current time step.

        """
        super().__init__()
        self.with_crn = with_crn
        self.sims_to_add = sims_to_add

    def setup(self, builder: Builder) -> None:
        self.register = builder.randomness.register_simulants
        self.randomness_init = builder.randomness.get_stream(
            "crn_init",
            initializes_crn_attributes=self.with_crn,
        )
        self.randomness_other = builder.randomness.get_stream("other")
        self.simulant_creator = builder.population.get_simulant_creator()

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        pass

    def on_time_step(self, event: Event) -> None:
        sims_to_add = next(self.sims_to_add)
        if sims_to_add > 0:
            self.simulant_creator(sims_to_add, {"sim_state": "time_step"})


class EntranceTimePopulation(BasePopulation):
    """Population that bases identity on entrance time and a random number"""

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        crn_attr = (1_000_000 * self.randomness_init.get_draw(index=pop_data.index)).astype(
            int
        )
        population = pd.DataFrame(
            {"crn_attr1": pop_data.creation_time, "crn_attr2": crn_attr},
            index=pop_data.index,
        )

        if self.with_crn:
            self.register(population)

        population["other_attr1"] = self.randomness_other.get_draw(
            pop_data.index,
            additional_key="attr1",
        )
        self.population_view.update(population)


class SequentialPopulation(BasePopulation):
    """
    Population that bases identity on the order simulants enter the simulation.

    NOTE: This population is not fully supported by the CRN system and is here to explicitly
    test and assert the expected failure cases.
    """

    def setup(self, builder: Builder) -> None:
        super().setup(builder)
        self.count = 0

    def on_initialize_simulants(self, pop_data: SimulantData):
        new_people = len(pop_data.index)

        population = pd.DataFrame(
            {
                "crn_attr1": pd.Timestamp("2020-01-01"),
                "crn_attr2": range(self.count, self.count + new_people),
            },
            index=pop_data.index,
        )

        if self.with_crn:
            self.register(population)

        population["other_attr1"] = self.randomness_other.get_draw(
            pop_data.index,
            additional_key="attr1",
        )
        self.population_view.update(population)
        self.count += new_people


@pytest.mark.parametrize(
    "pop_class, with_crn, sims_to_add",
    [
        pytest.param(EntranceTimePopulation, True, cycle([0])),
        pytest.param(EntranceTimePopulation, True, cycle([2])),
        pytest.param(EntranceTimePopulation, False, cycle([0])),
        pytest.param(EntranceTimePopulation, False, cycle([2])),
        pytest.param(SequentialPopulation, True, cycle([0])),
        pytest.param(SequentialPopulation, True, cycle([2])),
        pytest.param(SequentialPopulation, False, cycle([0])),
        pytest.param(SequentialPopulation, False, cycle([2])),
    ],
)
def test_multi_sim_basic_reproducibility_with_same_pop_growth(
    pop_class: Type,
    with_crn: bool,
    sims_to_add: cycle,
):
    if with_crn:
        configuration = {"randomness": {"key_columns": ["crn_attr1", "crn_attr2"]}}
    else:
        configuration = {}

    sim1 = InteractiveContext(
        components=[pop_class(with_crn=with_crn, sims_to_add=sims_to_add)],
        configuration=configuration,
    )
    sim2 = InteractiveContext(
        components=[pop_class(with_crn=with_crn, sims_to_add=sims_to_add)],
        configuration=configuration,
    )

    pop1 = sim1.get_population().set_index(["crn_attr1", "crn_attr2"])
    pop2 = sim2.get_population().set_index(["crn_attr1", "crn_attr2"])
    assert_frame_equal(pop1, pop2)

    for i in range(2):
        sim1.step()
        sim2.step()

    pop1 = sim1.get_population().set_index(["crn_attr1", "crn_attr2"])
    pop2 = sim2.get_population().set_index(["crn_attr1", "crn_attr2"])
    assert_frame_equal(pop1, pop2)


@pytest.mark.parametrize(
    "pop_class, with_crn",
    [
        pytest.param(EntranceTimePopulation, True),
        pytest.param(EntranceTimePopulation, False),
        pytest.param(SequentialPopulation, True, marks=pytest.mark.xfail),
        pytest.param(SequentialPopulation, False),
    ],
)
def test_multi_sim_reproducibility_with_different_pop_growth(pop_class: Type, with_crn: bool):
    if with_crn:
        configuration = {"randomness": {"key_columns": ["crn_attr1", "crn_attr2"]}}
    else:
        configuration = {}

    short, long = 1, 3
    sim1 = InteractiveContext(
        components=[pop_class(with_crn=with_crn, sims_to_add=cycle([short, short]))],
        configuration=configuration,
    )
    sim2 = InteractiveContext(
        components=[pop_class(with_crn=with_crn, sims_to_add=cycle([long, long]))],
        configuration=configuration,
    )

    pop1 = sim1.get_population().set_index(["crn_attr1", "crn_attr2"])
    pop2 = sim2.get_population().set_index(["crn_attr1", "crn_attr2"])
    initial_pop_size = len(pop1)
    assert_frame_equal(pop1, pop2)

    time_steps = 7
    for i in range(time_steps):
        sim1.step()
        sim2.step()

    pop1 = sim1.get_population().set_index(["crn_attr1", "crn_attr2"]).drop(columns="tracked")
    pop2 = sim2.get_population().set_index(["crn_attr1", "crn_attr2"]).drop(columns="tracked")

    if with_crn:
        overlap = pop1.index.intersection(pop2.index)
        assert len(overlap) > initial_pop_size
        assert_frame_equal(pop1.loc[overlap], pop2.loc[overlap])
    else:
        overlap = pop1.index[:initial_pop_size]
        assert_frame_equal(pop1.loc[overlap], pop2.loc[overlap])


class UnBrokenPopulation(BasePopulation):
    """
    CRN system used to fall over if the first CRN attribute is an int or float.

    This is now a regression testing class.
    """

    def on_initialize_simulants(self, pop_data: SimulantData):
        crn_attr = (1_000_000 * self.randomness_init.get_draw(index=pop_data.index)).astype(
            int
        )
        population = pd.DataFrame(
            {"crn_attr1": crn_attr, "crn_attr2": pop_data.creation_time},
            index=pop_data.index,
        )

        if self.with_crn:
            self.register(population)

        population["other_attr1"] = self.randomness_other.get_draw(
            pop_data.index,
            additional_key="attr1",
        )
        self.population_view.update(population)


@pytest.mark.parametrize(
    "with_crn, sims_to_add",
    [
        pytest.param(True, cycle([0])),
        pytest.param(True, cycle([1])),
        pytest.param(False, cycle([0])),
        pytest.param(False, cycle([1])),
    ],
)
def test_prior_failure_path_when_first_crn_attribute_not_datelike(
    with_crn: bool, sims_to_add: cycle
):
    if with_crn:
        configuration = {"randomness": {"key_columns": ["crn_attr1", "crn_attr2"]}}
    else:
        configuration = {}

    sim = InteractiveContext(
        components=[UnBrokenPopulation(with_crn=with_crn, sims_to_add=sims_to_add)],
        configuration=configuration,
    )

    for i in range(5):
        sim.step()



from typing import Any, Dict, List

import numpy as np
import pandas as pd

from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.population import SimulantData


class Location(Component):
    ##############
    # Properties #
    ##############
    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        return {
            "location": {
                "width": 1000,  # Width of our field
                "height": 1000,  # Height of our field
            }
        }

    @property
    def columns_created(self) -> List[str]:
        return ["x", "vx", "y", "vy"]

    #####################
    # Lifecycle methods #
    #####################

    def setup(self, builder: Builder) -> None:
        self.width = builder.configuration.location.width
        self.height = builder.configuration.location.height

    ########################
    # Event-driven methods #
    ########################

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        count = len(pop_data.index)
        # Start clustered in the center with small random velocities
        new_population = pd.DataFrame(
            {
                "x": self.width * (0.4 + 0.2 * np.random.random(count)),
                "y": self.height * (0.4 + 0.2 * np.random.random(count)),
                "vx": -0.5 + np.random.random(count),
                "vy": -0.5 + np.random.random(count),
            },
            index=pop_data.index,
        )
        self.population_view.update(new_population)



from typing import Dict, List, Optional

import pandas as pd

from vivarium import Component, InteractiveContext
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.population import SimulantData


class ColumnCreator(Component):
    @property
    def columns_created(self) -> List[str]:
        return ["test_column_1", "test_column_2", "test_column_3"]

    def setup(self, builder: Builder) -> None:
        builder.value.register_value_producer("pipeline_1", lambda x: x)
        builder.randomness.get_stream("stream_1")

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        initialization_data = pd.DataFrame(
            {column: 9 for column in self.columns_created}, index=pop_data.index
        )
        self.population_view.update(initialization_data)


class ColumnRequirer(Component):
    @property
    def columns_required(self) -> List[str]:
        return ["test_column_1", "test_column_2"]


class ColumnCreatorAndRequirer(Component):
    @property
    def columns_required(self) -> List[str]:
        return ["test_column_1", "test_column_2"]

    @property
    def columns_created(self) -> List[str]:
        return ["test_column_4"]

    @property
    def initialization_requirements(self) -> Dict[str, List[str]]:
        return {
            "requires_columns": ["test_column_2"],
            "requires_values": ["pipeline_1"],
            "requires_streams": ["stream_1"],
        }

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        initialization_data = pd.DataFrame({"test_column_4": 8}, index=pop_data.index)
        self.population_view.update(initialization_data)


class AllColumnsRequirer(Component):
    @property
    def columns_required(self) -> List[str]:
        return []


class FilteredPopulationView(ColumnRequirer):
    @property
    def population_view_query(self) -> Optional[str]:
        return "test_column_1 == 5"


class NoPopulationView(Component):
    pass


class Parameterized(Component):
    def __init__(self, p_one: str, p_two: int, p_three: str):
        super().__init__()
        self.p_one = p_one
        self.p_two = p_two
        self._p_three = p_three


class ParameterizedByComponent(Component):
    def __init__(self, param: Parameterized):
        super().__init__()
        self.param = param


class DefaultPriorities(Component):
    def on_post_setup(self, event: Event) -> None:
        pass

    def on_time_step_prepare(self, event: Event) -> None:
        pass

    def on_time_step(self, event: Event) -> None:
        pass

    def on_time_step_cleanup(self, event: Event) -> None:
        pass

    def on_collect_metrics(self, event: Event) -> None:
        pass

    def on_simulation_end(self, event: Event) -> None:
        pass


class CustomPriorities(DefaultPriorities):
    @property
    def post_setup_priority(self) -> int:
        return 8

    @property
    def time_step_prepare_priority(self) -> int:
        return 7

    @property
    def time_step_priority(self) -> int:
        return 2

    @property
    def time_step_cleanup_priority(self) -> int:
        return 3

    @property
    def collect_metrics_priority(self) -> int:
        return 6

    @property
    def simulation_end_priority(self) -> int:
        return 1


def test_unique_component_has_correct_repr():
    component = NoPopulationView()

    # Assert component has the correct repr
    assert component.__repr__() == "NoPopulationView()"


def test_parameterized_component_has_repr_that_incorporates_arguments():
    component = Parameterized("some_value", 5, "another_value")

    # Assert component has the correct repr
    assert component.__repr__() == "Parameterized(p_one=some_value, p_two=5)"


def test_component_by_other_component_has_repr_that_incorporates_arguments():
    arg = Parameterized("some_value", 5, "another_value")
    component = ParameterizedByComponent(arg)

    # Assert component has the correct repr
    expected_repr = "ParameterizedByComponent(param=Parameterized(p_one=some_value, p_two=5))"
    assert component.__repr__() == expected_repr


def test_component_with_no_arguments_has_correct_name():
    component = NoPopulationView()

    # Assert component has the correct name
    assert component.name == "no_population_view"


def test_parameterized_component_has_name_that_incorporates_arguments():
    component = Parameterized("some_value", 5, "another_value")

    # Assert component has the correct name
    assert component.name == "parameterized.some_value.5"


def test_component_by_other_component_has_name_that_incorporates_arguments():
    arg = Parameterized("some_value", 5, "another_value")
    component = ParameterizedByComponent(arg)

    # Assert component has the correct repr
    expected = "parameterized_by_component.'parameterized.some_value.5'"
    assert component.name == expected


def test_component_that_creates_columns_population_view():
    component = ColumnCreator()
    InteractiveContext(components=[component])

    # Assert population view is set and has the correct columns
    assert component.population_view is not None
    assert set(component.population_view.columns) == set(component.columns_created)


def test_component_that_requires_columns_population_view():
    component = ColumnRequirer()
    InteractiveContext(components=[ColumnCreator(), component])

    # Assert population view is set and has the correct columns
    assert component.population_view is not None
    assert set(component.population_view.columns) == set(component.columns_required)


def test_component_that_creates_and_requires_columns_population_view():
    component = ColumnCreatorAndRequirer()
    InteractiveContext(components=[ColumnCreator(), component])

    # Assert population view is set and has the correct columns
    expected_columns = component.columns_required + component.columns_created

    assert component.population_view is not None
    assert set(component.population_view.columns) == set(expected_columns)


def test_component_with_initialization_requirements():
    component = ColumnCreatorAndRequirer()
    simulation = InteractiveContext(components=[ColumnCreator(), component])

    # Assert required resources have been recorded by the ResourceManager
    component_dependencies_list = [
        r.dependencies
        # get all resources in the dependency graph
        for r in simulation._resource.sorted_nodes
        # if the producer is an instance method
        if hasattr(r.producer, "__self__")
        # and is a method of ColumnCreatorAndRequirer
        and isinstance(r.producer.__self__, ColumnCreatorAndRequirer)
    ]
    assert len(component_dependencies_list) == 1
    component_dependencies = component_dependencies_list[0]

    assert "value.pipeline_1" in component_dependencies
    assert "column.test_column_2" in component_dependencies
    assert "stream.stream_1" in component_dependencies


def test_component_that_requires_all_columns_population_view():
    component = AllColumnsRequirer()
    simulation = InteractiveContext(
        components=[ColumnCreator(), ColumnCreatorAndRequirer(), component]
    )
    population = simulation.get_population()

    # Assert population view is set and has the correct columns
    expected_columns = population.columns

    assert component.population_view is not None
    assert set(component.population_view.columns) == set(expected_columns)


def test_component_with_filtered_population_view():
    component = FilteredPopulationView()
    InteractiveContext(components=[ColumnCreator(), component])

    # Assert population view is being filtered using the desired query
    assert component.population_view.query == "test_column_1 == 5 and tracked == True"


def test_component_with_no_population_view():
    component = NoPopulationView()
    InteractiveContext(components=[ColumnCreator(), component])

    # Assert population view is not set
    assert component.population_view is None


def test_component_initializer_is_not_registered_if_not_defined():
    component = NoPopulationView()
    simulation = InteractiveContext(components=[component])

    # Assert that simulant initializer has been registered
    assert component.on_initialize_simulants not in simulation._resource


def test_component_initializer_is_registered_and_called_if_defined():
    component = ColumnCreator()
    simulation = InteractiveContext(components=[component])
    population = simulation.get_population()
    expected_pop_view = pd.DataFrame(
        {column: 9 for column in component.columns_created}, index=population.index
    )

    # Assert that simulant initializer has been registered
    assert component.on_initialize_simulants in simulation._resource
    # and that created columns are correctly initialized
    pd.testing.assert_frame_equal(population[component.columns_created], expected_pop_view)


def test_listeners_are_not_registered_if_not_defined():
    component = NoPopulationView()
    simulation = InteractiveContext(components=[component])

    post_setup_methods = simulation.get_listeners("post_setup")
    time_step_prepare_methods = simulation.get_listeners("time_step__prepare")
    time_step_methods = simulation.get_listeners("time_step")
    time_step_cleanup_methods = simulation.get_listeners("time_step__cleanup")
    collect_metrics_methods = simulation.get_listeners("collect_metrics")
    simulation_end_methods = simulation.get_listeners("simulation_end")

    for i in range(10):
        assert component.on_post_setup not in set(post_setup_methods.get(i, []))
        assert component.on_time_step_prepare not in set(time_step_prepare_methods.get(i, []))
        assert component.on_time_step not in set(time_step_methods.get(i, []))
        assert component.on_time_step_cleanup not in set(time_step_cleanup_methods.get(i, []))
        assert component.on_collect_metrics not in set(collect_metrics_methods.get(i, []))
        assert component.on_simulation_end not in set(simulation_end_methods.get(i, []))


def test_listeners_are_registered_if_defined():
    component = DefaultPriorities()
    simulation = InteractiveContext(components=[component])

    post_setup_methods = simulation.get_listeners("post_setup")
    time_step_prepare_methods = simulation.get_listeners("time_step__prepare")
    time_step_methods = simulation.get_listeners("time_step")
    time_step_cleanup_methods = simulation.get_listeners("time_step__cleanup")
    collect_metrics_methods = simulation.get_listeners("collect_metrics")
    simulation_end_methods = simulation.get_listeners("simulation_end")

    assert component.on_post_setup in set(post_setup_methods.get(5, []))
    assert component.on_time_step_prepare in set(time_step_prepare_methods.get(5, []))
    assert component.on_time_step in set(time_step_methods.get(5, []))
    assert component.on_time_step_cleanup in set(time_step_cleanup_methods.get(5, []))
    assert component.on_collect_metrics in set(collect_metrics_methods.get(5, []))
    assert component.on_simulation_end in set(simulation_end_methods.get(5, []))


def test_listeners_are_registered_at_custom_priorities():
    component = CustomPriorities()
    simulation = InteractiveContext(components=[component])

    post_setup_methods = simulation.get_listeners("post_setup")
    time_step_prepare_methods = simulation.get_listeners("time_step__prepare")
    time_step_methods = simulation.get_listeners("time_step")
    time_step_cleanup_methods = simulation.get_listeners("time_step__cleanup")
    collect_metrics_methods = simulation.get_listeners("collect_metrics")
    simulation_end_methods = simulation.get_listeners("simulation_end")

    assert component.on_post_setup not in set(post_setup_methods.get(5, []))
    assert component.on_time_step_prepare not in set(time_step_prepare_methods.get(5, []))
    assert component.on_time_step not in set(time_step_methods.get(5, []))
    assert component.on_time_step_cleanup not in set(time_step_cleanup_methods.get(5, []))
    assert component.on_collect_metrics not in set(collect_metrics_methods.get(5, []))
    assert component.on_simulation_end not in set(simulation_end_methods.get(5, []))

    assert component.on_post_setup in set(post_setup_methods.get(8, []))
    assert component.on_time_step_prepare in set(time_step_prepare_methods.get(7, []))
    assert component.on_time_step in set(time_step_methods.get(2, []))
    assert component.on_time_step_cleanup in set(time_step_cleanup_methods.get(3, []))
    assert component.on_collect_metrics in set(collect_metrics_methods.get(6, []))
    assert component.on_simulation_end in set(simulation_end_methods.get(1, []))



"""
=========================
Randomness System Manager
=========================
"""
import pandas as pd

from vivarium.framework.randomness.exceptions import RandomnessError
from vivarium.framework.randomness.index_map import IndexMap
from vivarium.framework.randomness.stream import RandomnessStream, get_hash
from vivarium.manager import Manager


class RandomnessManager(Manager):
    """Access point for common random number generation."""

    configuration_defaults = {
        "randomness": {
            "map_size": 1_000_000,
            "key_columns": [],
            "random_seed": 0,
            "additional_seed": None,
        }
    }

    def __init__(self):
        self._seed = None
        self._clock = None
        self._key_columns = None
        self._key_mapping = None
        self._decision_points = dict()

    @property
    def name(self):
        return "randomness_manager"

    def setup(self, builder):
        self._seed = str(builder.configuration.randomness.random_seed)
        if builder.configuration.randomness.additional_seed is not None:
            self._seed += str(builder.configuration.randomness.additional_seed)
        self._clock = builder.time.clock()
        self._key_columns = builder.configuration.randomness.key_columns

        map_size = builder.configuration.randomness.map_size
        pop_size = builder.configuration.population.population_size
        map_size = max(map_size, 10 * pop_size)
        self._key_mapping = IndexMap(self._key_columns, map_size)

        self.resources = builder.resources
        self._add_constraint = builder.lifecycle.add_constraint
        self._add_constraint(self.get_seed, restrict_during=["initialization"])
        self._add_constraint(self.get_randomness_stream, allow_during=["setup"])
        self._add_constraint(
            self.register_simulants,
            restrict_during=[
                "initialization",
                "setup",
                "post_setup",
                "simulation_end",
                "report",
            ],
        )

    def get_randomness_stream(
        self, decision_point: str, initializes_crn_attributes: bool = False
    ) -> RandomnessStream:
        """Provides a new source of random numbers for the given decision point.

        Parameters
        ----------
        decision_point
            A unique identifier for a stream of random numbers.  Typically
            represents a decision that needs to be made each time step like
            'moves_left' or 'gets_disease'.
        initializes_crn_attributes
            A flag indicating whether this stream is used to generate key
            initialization information that will be used to identify simulants
            in the Common Random Number framework. These streams cannot be
            copied and should only be used to generate the state table columns
            specified in ``builder.configuration.randomness.key_columns``.

        Raises
        ------
        RandomnessError
            If another location in the simulation has already created a randomness stream
            with the same identifier.

        """
        stream = self._get_randomness_stream(decision_point, initializes_crn_attributes)
        if not initializes_crn_attributes:
            # We need the key columns to be created before this stream can be called.
            self.resources.add_resources(
                "stream",
                [decision_point],
                stream,
                [f"column.{name}" for name in self._key_columns],
            )
        self._add_constraint(
            stream.get_draw, restrict_during=["initialization", "setup", "post_setup"]
        )
        self._add_constraint(
            stream.filter_for_probability,
            restrict_during=["initialization", "setup", "post_setup"],
        )
        self._add_constraint(
            stream.filter_for_rate, restrict_during=["initialization", "setup", "post_setup"]
        )
        self._add_constraint(
            stream.choice, restrict_during=["initialization", "setup", "post_setup"]
        )

        return stream

    def _get_randomness_stream(
        self, decision_point: str, initializes_crn_attributes: bool = False
    ) -> RandomnessStream:
        if decision_point in self._decision_points:
            raise RandomnessError(
                f"Two separate places are attempting to create "
                f"the same randomness stream for {decision_point}"
            )
        stream = RandomnessStream(
            key=decision_point,
            clock=self._clock,
            seed=self._seed,
            index_map=self._key_mapping,
            initializes_crn_attributes=initializes_crn_attributes,
        )
        self._decision_points[decision_point] = stream
        return stream

    def get_seed(self, decision_point: str) -> int:
        """Get a randomly generated seed for use with external randomness tools.

        Parameters
        ----------
        decision_point
            A unique identifier for a stream of random numbers.  Typically
            represents a decision that needs to be made each time step like
            'moves_left' or 'gets_disease'.

        Returns
        -------
        int
            A seed for a random number generation that is linked to Vivarium's
            common random number framework.

        """
        return get_hash("_".join([decision_point, str(self._clock()), str(self._seed)]))

    def register_simulants(self, simulants: pd.DataFrame):
        """Adds new simulants to the randomness mapping.

        Parameters
        ----------
        simulants
            A table with state data representing the new simulants.  Each
            simulant should pass through this function exactly once.

        Raises
        ------
        RandomnessError
            If the provided table does not contain all key columns specified
            in the configuration.

        """
        if not all(k in simulants.columns for k in self._key_columns):
            raise RandomnessError(
                "The simulants dataframe does not have all specified key_columns."
            )
        self._key_mapping.update(simulants.loc[:, self._key_columns], self._clock())

    def __str__(self):
        return "RandomnessManager()"

    def __repr__(self) -> str:
        return f"RandomnessManager(seed={self._seed}, key_columns={self._key_columns})"


class RandomnessInterface:
    def __init__(self, manager: RandomnessManager):
        self._manager = manager

    def get_stream(
        self, decision_point: str, initializes_crn_attributes: bool = False
    ) -> RandomnessStream:
        """Provides a new source of random numbers for the given decision point.

        ``vivarium`` provides a framework for Common Random Numbers which
        allows for variance reduction when modeling counter-factual scenarios.
        Users interested in causal analysis and comparisons between simulation
        scenarios should be careful to use randomness streams provided by the
        framework wherever randomness is employed.

        Parameters
        ----------
        decision_point
            A unique identifier for a stream of random numbers.  Typically
            represents a decision that needs to be made each time step like
            'moves_left' or 'gets_disease'.
        initializes_crn_attributes
            A flag indicating whether this stream is used to generate key
            initialization information that will be used to identify simulants
            in the Common Random Number framework. These streams cannot be
            copied and should only be used to generate the state table columns
            specified in ``builder.configuration.randomness.key_columns``.

        Returns
        -------
        RandomnessStream
            An entry point into the Common Random Number generation framework.
            The stream provides vectorized access to random numbers and a few
            other utilities.

        """
        return self._manager.get_randomness_stream(decision_point, initializes_crn_attributes)

    def get_seed(self, decision_point: str) -> int:
        """Get a randomly generated seed for use with external randomness tools.

        Parameters
        ----------
        decision_point :
            A unique identifier for a stream of random numbers.  Typically
            represents a decision that needs to be made each time step like
            'moves_left' or 'gets_disease'.

        Returns
        -------
        int
            A seed for a random number generation that is linked to Vivarium's
            common random number framework.

        """
        return self._manager.get_seed(decision_point)

    def register_simulants(self, simulants: pd.DataFrame):
        """Registers simulants with the Common Random Number Framework.

        Parameters
        ----------
        simulants
            A section of the state table with new simulants and at least the
            columns specified in
            ``builder.configuration.randomness.key_columns``.  This function
            should be called as soon as the key columns are generated.

        """
        self._manager.register_simulants(simulants)
